/********************************************
 *---- Assignment 1 -------
 * Name  : Arvind S Krishna
 * Id    : 90086152
 * Email : krishnaa@uci.edu
********************************************/

/**
 * Program Execution:
 * -------------------
 * This program takes in the number of threads to span as a command line argument
 * and creates the corresponding number of pthreads, Each of the threads do a busy
 * wait for the time period of 1 sec specified in the wait_time external variable.
 * To run the program : a.out <number of threads>.
 * 
 * External Varibles:
 * ------------------
 * wait_period: Is the time that the threads have to do busy wait before a print. E.g
 *              The thread id is printed after every one sec. So in this case 
 *              wait_period is 1 sec. The default value is set to 1 sec.
 *
 * print_itretaions: Indicates the number of times the Thread id will be printed after
 *                   completion of every wait_period. E.g In our case the thread
 *                   id is printed 5 times after the thread completes waiting for 1 sec.
 *
 * calibration_iterations: The number of times the time required to complete one second
 *                         by the time_wasting fn is computed.
 * 
 * iterations : Contains the number of iterations required by the time_wasting_fn to
 *              complete the time specified in wait_period.
 *
 * Function Decalrations:
 * ----------------------
 * perform_work: This is the point of concurrent execution of the pthreads. This is 
 *               also the busy wait for the Threads. This function is RE-ENTRANT as it
 *               does not modify any globals or statics and creates no RACE Conditions.
 *
 * calibration_loop: This function calibrates the number of loop iterations that are 
 *                   required for the time wasting function to complete the time 
 *                   specified in the wait_time variable. The number of iterations
 *                   this is done is specified in the calibration_iterations varible.
 *
 * time_wasting_function: This function is used to compute some junk value that takes up
 *                       some time. This function is called N times so that the total 
 *                       time used = time specified in the wait_time varible.
 *
 * print_help: Is a helper function that is used for printing the the right way of 
 *             executing this program. Can be invoked with also --help option. 
 */



/*--- Header files ---*/
#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <time.h>
#include <sys/time.h>

/*-- Function Declarations --*/
void* perform_work (void*);
long calibration_loop (void);
void time_wasting_function (void);
void print_help (void);

/***** External Variables ******************************/
long iterations =0;

/* Number of iterations that the calibration loop will execute */
int calibration_iterations = 10;
 
/* Wait period is one sec by default */
int wait_period = 1; 

/* The number of times the Thread status is printed after every wait_period */
int print_iterations = 5;
/*******************************************************/

int main( int argc, char* argv[])
{
  pthread_t *threads[];
  int num_threads, counter, error, temp_result =0;

  /* This indicates the number of secs the time_wasting fn should consume
   * by default we set this to 1 sec
   */
  long wait_period; 

  pthread_attr_t attr;

  /*-- Command Line option where the user enters "a.out --help" --*/
  if ( argc == 2 && strcmp (argv[1], "--help") == 0)
  {
    print_help ();
    exit(0);
  }

  /*---  Validate the input ---*/
  if (argc != 2)
  {
    printf ("\n Incorrect Argument Exception: more than one/no arguments given in the command line\n");
    print_help ();
    exit(-1);
  }
  
  /*---The number of threads and wait_period is got from the command line */
  num_threads = atoi (argv[1]);
  wait_period = 1; /* The time to wait is set to one sec by default */

  /*--- Check if they are positive integers!! ---*/
  if (num_threads < 0)
  {
     printf ("IllegalArgumentException: Negative Arguments passed for the number");
     printf ("of threads\n");
     print_help ();
     exit(-1);
  } 

  if (num_threads ==0)
  {
    printf ("\nIllegalArgumentException: for any of the following reasons \n");
    printf ("1: 0 passed as the number of threads to be created \n");
    printf ("2: Alpha Numeric Input given for number of threads \n");
    print_help();
    exit (-1);
  }


  /*--- Run the loop to calibrate the time to wait ---*/
   printf ("\n\nCalculating the number of iterations needed for the time_wasting_function() \n");
   printf ("to complete 1 second \n");
   printf ("**************************************************************************\n");
   
   /* --- One dry run to eliminate exceptional cases --*/
   printf ("Dry Run: Not counted :");
   calibration_loop();
   
   for (counter =1; counter <= calibration_iterations; counter++)
   {
      printf ("Run %d : ", counter);
      temp_result += calibration_loop ();
   }

   /* --- Set the Avg number of iterations required to completed one second --*/
   iterations = temp_result / calibration_iterations;

   printf ("The average number of iterations required to complete one sec = %d\n", iterations);
   printf ("**************************************************************************\n");
  
   /*--- Initiazile the environment and also enable system wide contention --*/
   pthread_attr_init(&attr); /* initialize attr with default attributes */
   pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM); /* system-wide contention */
  
   for(counter =0; counter < num_threads; counter++)
   {
     /*--- Allocate memory to the pthreads ---*/
     threads[counter] = (pthread_t*)malloc(sizeof(pthread_t));

     /*-- Create the p_thread  ---*/
     error = pthread_create (threads[counter], &attr, 
	 	            perform_work, (void*) threads[counter]);
     if (error != 0)
     {
       printf ("Error:Thread creation has failed");
       exit(-1);
     }
     printf ("Thread:%d starting \n", *threads[counter]);

   }
 
   /* --- Join the threads ----*/ 
   for(counter =0; counter < num_threads; counter++)
     pthread_join (*threads[counter], NULL);
}

/*
 * perform_work() function is the point of concurrent execution of the 
 * threads. This function <b> wastes </b> time by doing some mathematical
 * computation based on the time_wasting() function. 
 * perform_work is also re-entrant and can be used for all the 
 * threads.
 */
 
 void* perform_work (void* ptr)
 { 
   struct timeval current_time, loop_time;
   int counter=0, outer, number = *( (int*)ptr );

   /*--- Do the busy waiting part --*/
   for (outer=1; outer <= print_iterations; outer++)
   {
     for (counter =1; counter <= iterations; counter++)
       time_wasting_function ();

     /* --- After every one second print the Thread id ---*/
     printf ("T%d-%d ", number, outer);
     fflush (stdout);
   }

   printf ("Thread %d exiting\n", number);
 }
 
 /*
  * The calibration loop is run apriori *calibration_iteration* times to 
  * correctly measure the number of iterations that have to elapse for the 
  * time_wasting() to complete one second.  
  */

 long calibration_loop (void)
 {
   struct timeval current_time, loop_time;
   long counter =1;
   
   /*--- Get the time now and run the time wasting loop to complete one sec --*/
   gettimeofday (&current_time,NULL);
   do
   {
       time_wasting_function ();
       counter++;
       gettimeofday (&loop_time,NULL);
   }while (loop_time.tv_sec <= current_time.tv_sec + wait_period);
   
   printf ("No of loop iterations: %d \n", counter);
   
   return counter;
 }

/*
 * time_wasting() function is the computation performed by each thread as a part of the
 * busy wait. This is a junk computation and takes a fixed 
 * amount of time.
 * 
 * Basically number_of_iterations * time for the this function = wait_period
 */
 
void time_wasting_function (void)
{
   int counter =0;
   counter = (counter +1)/ 100 * counter + 3;
}

/*
 * print_help () function contains all the help instructions that are necessary
 * to run the program correctly. This function is invoked when ever there is an 
 * incorrect input from the user.
 */

void print_help (void)
{
    printf ("\n\nTo run this program do the following: \n");
    printf ("---------------------------------------\n");
    printf ("1:Enter the number of threads that the program should span \n");
    printf (" The Number of threads should be a positive Integer: \n");
    printf (" Entering Floating point Numbers: Integer Value will be taken\n");
    printf ("\nAt the command prompt do the following\n");
    printf ("./a.out Number_of_Threads <enter> \n\n\n");
}
