/********************************************
 *---- Assignment 2 -------
 * Name  : Arvind S Krishna
 * Id    : 90086152
 * Email : krishnaa@uci.edu
********************************************/

/**
 * Program Execution:
 * -------------------
 * This program takes in the number of threads to span as a command line argument
 * and creates the corresponding number of pthreads, Each of the threads do a busy
 * wait for the time period of 1 sec specified in the wait_time external variable.
 * To run the program : a.out <number of threads>.
 * 
 * External Varibles:
 * ------------------
 * wait_period: Is the time that the threads have to do busy wait before a print. E.g
 *              The thread id is printed after every one sec. So in this case 
 *              wait_period is 1 sec. The default value is set to 1 sec.
 *
 * print_itretaions: Indicates the number of times the Thread id will be printed after
 *                   completion of every wait_period. E.g In our case the thread
 *                   id is printed 5 times after the thread completes waiting for 1 sec.
 *
 * calibration_iterations: The number of times the time required to complete one second
 *                         by the time_wasting fn is computed.
 * 
 * iterations : Contains the number of iterations required by the time_wasting_fn to
 *              complete the time specified in wait_period.
 *
 * Function Decalrations:
 * ----------------------
 * perform_work: This is the point of concurrent execution of the pthreads. This is 
 *               also the busy wait for the Threads. This function is RE-ENTRANT as it
 *               does not modify any globals or statics and creates no RACE Conditions.
 *
 *
 * print_help: Is a helper function that is used for printing the the right way of 
 *             executing this program. Can be invoked with also --help option. 
 */



/*--- System Header files  and other include files---*/
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <time.h>
#include <ctype.h>
#include <pthread.h>

#include "Time.h"
#include "Mutex.h"

#define _MULTI_THREADED

/*-- Function Declarations --*/
void* perform_work (void*);
void print_help (void);
extern void set_delay (int);
extern void initialize_mutex (void);
extern int get_thread_id (void);
extern void destroy_mutex (void);
extern long calibration_loop (int);

/***** External Variables ******************************/
long iterations =0;
int calibration_iterations = 5;
int print_iterations = 5;
/*******************************************************/

int main( int argc, char* argv[])
{
  pthread_t *threads[];
  pthread_attr_t *attr[];

  int num_threads, counter, error, delay, temp_result =0;

  /* This indicates the number of secs the time_wasting fn should consume
   * by default we set this to 1 sec
   */
  long wait_period; 

  /*-- Command Line option where the user enters "a.out --help" --*/
  if ( argc == 2 && strcmp (argv[1], "--help") == 0)
  {
    print_help ();
    exit(0);
  }

  /*---  Validate the input ---*/
  if (argc != 3)
  {
    printf ("\n Incorrect Argument Exception: Incorrect Number of arguments given in the command line\n");
    print_help ();
    exit(-1);
  }
  
  /*---The number of threads and wait_period is got from the command line */
  num_threads = atoi (argv[1]);
  delay = atoi (argv[2]);

  if (strcmp (argv[2], "0") != 0 &&  delay == 0)
  {
      printf ("Illegal Argument Exception: The second argument is not a number!!\n");
      print_help ();
      exit (-1);
  }

  /* The time to wait is set to one sec by default */
  wait_period = 1;   

  /*--- Check if they are positive integers!! ---*/
  if (num_threads < 0 || delay < 0)
  {
     printf ("IllegalArgumentException: Negative Numbers passed as argument");
     print_help ();
     exit(-1);
  } 

  if (num_threads ==0)
  {
    printf ("\nIllegalArgumentException: for any of the following reasons \n");
    printf ("1: 0 passed as the number of threads to be created \n");
    printf ("2: Alpha Numeric Input given for number of threads \n");
    print_help();
    exit (-1);
  }

  if (delay > 1)
  {
    printf ("\n\nThe second argument should be either 0 or 1 only\n");
    print_help ();
    exit (-1);
  }

   set_delay (delay); 

   /*--- Run the loop to calibrate the time to wait ---*/
   printf ("\n\nCalculating the number of iterations needed for the time_wasting _function() \n");
   printf ("to complete 1 second \n");
   printf ("**************************************************************************\n");

   printf ("Dry Run: Not counted :");
   calibration_loop (wait_period);

   for (counter =1; counter <= calibration_iterations; counter++)
   {
      printf ("Run %d : ", counter);
      temp_result += calibration_loop (wait_period);
   }

   /* --- Set the Avg number of iterations required to completed one second --*/
   iterations = temp_result / calibration_iterations;

   printf ("The average number of iterations required to complete one sec = %d\n ", iterations);

   printf ("*************************************************************************\n");

 
   /* Initialize the Mutex Lock */
    initialize_mutex (); 
 
   for(counter =0; counter < num_threads; counter++)
   {
     /*--- Allocate memory to the pthreads ---*/
     threads[counter]   = (pthread_t*)malloc(sizeof(pthread_t));
     attr[counter]  = (pthread_attr_t*)malloc (sizeof(pthread_attr_t));

     pthread_attr_init (attr[counter]); /* initialize attr with default attributes */
     pthread_attr_setscope (attr[counter], PTHREAD_SCOPE_SYSTEM); /* system-wide contention */

     /*-- Create the p_thread  ---*/
     error = pthread_create (threads[counter], attr[counter], 
	 	            perform_work, (void*) &delay);
     if (error != 0)
     {
       printf ("Error:Thread creation has failed %d: %d \n", error,counter);
       switch (error)
       {
          case ENOMEM: printf ("Not Enough memory in the System\n");
                       break;
          case EINVAL: printf ("value specified in the attr is invalid\n");
                       break;
          case EPERM:  printf ("The caller does not have the enough permission to create threads \n");
                       break;
	  case EAGAIN: printf ("Failed due to EAGAIN!!\n");
	               break;

	  default:     printf ("The reason why it failed is unknown\n");
       }
       exit(-1);
     }
     
   }
 
   /* --- Join the threads ---- */
   for(counter =0; counter < num_threads; counter++)
     pthread_join (*threads[counter], NULL); 

   /* ---- Destroy the Lock ---*/
    destroy_mutex (); 
   
}

/*
 * perform_work() function is the point of concurrent execution of the 
 * threads. This function <b> wastes </b> time by doing some mathematical
 * computation based on the time_wasting() function. 
 * perform_work is also re-entrant and can be used for all the 
 * threads.
 */
 
 void* perform_work (void* ptr)
 { 
   struct timeval current_time, loop_time;
   int counter=0, outer, id; 
   extern int delay;

   /* Get the Thread id of the thread */
   id = get_thread_id (); 
   
   printf ("Thread:%d starting \n", id);

   /*--- Do the busy waiting part --*/
   for (outer=1; outer <= print_iterations; outer++)
   {
     for (counter =1; counter <= iterations; counter++)
       time_wasting_function ();

     /* --- After every one second print the Thread id ---*/
     printf ("T%d-%d ", id, outer);
     fflush (stdout);
   }

   printf ("Thread %d exiting\n", id);
 }
 
/*
 * print_help () function contains all the help instructions that are necessary
 * to run the program correctly. This function is invoked when ever there is an 
 * incorrect input from the user.
 */

void print_help (void)
{
    printf ("\n\nTo run this program do the following: \n");
    printf ("---------------------------------------\n");
    printf ("1:Enter the number of threads that the program should span \n");
    printf (" The Number of threads should be a positive Integer: \n");
    printf ("2:Enter 1 or 0 to indicate whether there should be a delay loop\n");
    printf (" Entering Floating point Numbers: Integer Value will be taken in both the cases\n");
    printf ("\nAt the command prompt do the following\n");
    printf ("./a.out Number_of_Threads <0 or 1> <enter> \n\n\n");
}
