/********************************************
 *---- Assignment 3 -------
 * Name  : Arvind S Krishna
 * Id    : 90086152
 * Email : krishnaa@uci.edu
********************************************/

/*--- System Header files  and other include files---*/
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <ctype.h>
#include <pthread.h>
#include <mqueue.h>
#include <string.h>

unsigned long PMODE =  0666;
unsigned int  PRTY = 0;
long MSG_LEN = 3;

int num_msg, msg_size, pairs;

/* --Names for the message queues ---*/
char ** names;

/*-- Function Declarations --*/
void* produce (void*);
void* consume (void*);
void print_help (void);
void create_message_queues (void);
void close_message_queues (int);

int main( int argc, char* argv[])
{
  pthread_t *threads[];
  pthread_attr_t *attr[];

  int counter, error, id, temp;
  char s[1], *prefix = "/kraa";
  int *queue_no;

  /*-- Command Line option where the user enters "a.out --help" --*/
  if ( argc == 2 && strcmp (argv[1], "--help") == 0)
  {
    print_help ();
    exit(0);
  }

  /*---  Validate the input ---*/
  if (argc != 4)
  {
    printf ("\n Incorrect Argument Exception: Incorrect Number of arguments given in the command line\n");
    print_help ();
    exit(-1);
  }
  
  /*---The number of threads and wait_period is got from the command line */
  pairs = atoi (argv[1]);
  num_msg= atoi (argv[2]);
  msg_size = atoi (argv[3]);

  /*--- Check if they are positive integers!! ---*/
  if (pairs <= 0 || msg_size <= 0 || num_msg <= 0)
  {
     printf ("IllegalArgumentException: Negative Numbers passed as argument");
     print_help ();
     exit(-1);
  }

   queue_no = (int*) malloc(pairs* sizeof(int));
   names   = (char**) malloc(pairs * sizeof(char*));
   
   for (counter=0; counter < pairs; counter++)
   {
      names [counter] = (char*) malloc(sizeof(prefix));
      strcpy (names[counter], prefix);
      temp = counter;
      do{
         s [0] = temp % 10 + '0';
         strcat (names[counter], s);
         temp = temp/10;
      }while ( temp > 0);
   }

   for(counter =0; counter < 2*pairs; counter++)
   {
     /*--- Allocate memory to the pthreads ---*/
     threads[counter]   = (pthread_t*)malloc(sizeof(pthread_t));
     attr[counter]  = (pthread_attr_t*)malloc (sizeof(pthread_attr_t));

     pthread_attr_init (attr[counter]); /* initialize attr with default attributes */
     pthread_attr_setscope (attr[counter], PTHREAD_SCOPE_SYSTEM); 

     /*-- Create the p_thread  ---*/
     if ((counter/pairs) == 0) 
     {
       queue_no[counter] = counter;
       error = pthread_create (threads[counter], attr[counter], 
	 	               produce, (void*) &queue_no[counter]);
     }
     else
       error = pthread_create (threads[counter], attr[counter], 
	 	               consume, (void*) &queue_no[counter%pairs]);
       
     if (error != 0)
     {
       printf ("Error:Thread creation has failed %d: %d \n", error,counter);
       switch (error)
       {
          case ENOMEM: printf ("Not Enough memory in the System\n");
                       break;
          case EINVAL: printf ("value specified in the attr is invalid\n");
                       break;
          case EPERM:  printf ("The caller does not have the enough permission to create threads \n");
                       break;
	  case EAGAIN: printf ("Failed due to EAGAIN!!\n");
	               break;

	  default:     printf ("The reason why it failed is unknown\n");
       }
       exit(-1);
     }
     
   }

   /* --- Join the threads ---- */
   for(counter =0; counter < 2*pairs; counter++)
     pthread_join (*threads[counter], NULL); 

    close_message_queues (pairs); 
    
    free ((void*) queue_no);
    for( counter=0; counter < pairs; counter++)
        free ((void*)names[counter]);

    free ((void**)names);
}

 
 void* produce (void* ptr)
 { 
   int i, status, index;
   mqd_t fd; 
   struct mq_attr attr;

   /* --- Buffer for copying of the data into the buffers --*/
   char * buffer = (char *) malloc (msg_size * sizeof(char));

   /* -- index into the message queue indexes --*/
   index = * (int*)ptr;
   
   printf ("Producer Thread %d starting: \n", index); 

   attr.mq_msgsize  = (long)msg_size;
   attr.mq_maxmsg   = MSG_LEN;

   fd = mq_open (names[index], O_CREAT|O_WRONLY,PMODE,&attr);

   if (fd == (mqd_t) -1) {
      perror ("Producer open failed");
      exit (0);
   }
   

   for (i=0; i < msg_size; i++)
       buffer [i] = 'A'; 

   for (i=0; i < num_msg; i++)
   {
     status = mq_send (fd, buffer, msg_size, PRTY);
     if (status == -1)
     {
       perror ("send failed \n");
       mq_close (fd);
       return;
     }
   }

   printf ("Producer %d exiting \n", index); 
   free ((void*) buffer);
    
 }

 void* consume (void* ptr)
 { 
   int i, num_bytes=0, index, temp;
   mqd_t fd;
   struct mq_attr q_attr;
   char *buffer; 
   
   index = * (int*) ptr; 

   q_attr.mq_msgsize = (long)msg_size;
   q_attr.mq_maxmsg  =  MSG_LEN;

   fd = mq_open (names[index], O_CREAT|O_RDONLY,PMODE,&q_attr);

   if (fd == (mqd_t) -1) {
      perror ("Consumer open failed");
      exit (0);
   }

   buffer = (char *)malloc (q_attr.mq_msgsize * sizeof(char));

   printf ("Consumer Thread %d starting: \n",index); 
   for (i =0; i < num_msg; i++)
   {
         temp = mq_receive (fd, buffer, q_attr.mq_msgsize, &PRTY);
         if (temp == -1)
         {
             perror("mq_receive failure on messge_queue");
             mq_close (fd);
             return;
          }
   }
   printf ("Consumer %d exiting \n", index);
   close (fd);
   free ((void*) buffer);
 }
 
 
/*
 * print_help () function contains all the help instructions that are necessary
 * to run the program correctly. This function is invoked when ever there is an 
 * incorrect input from the user.
 */

void print_help (void)
{
    printf ("\n\nTo run this program do the following: \n");
    printf ("---------------------------------------\n");
    printf ("1:Enter the number of pairs \n");
    printf ("2:The Number of messages that have to be sent: \n");
    printf ("3: Size of the messages that have to be sent\n");
    printf (" Entering Floating point Numbers: Integer Value will be taken in both the cases\n");
    printf ("\nAt the command prompt do the following\n");
}

void close_message_queues (int num_queue)
{
   int i;

   for (i=0; i < pairs; i++)
   {
     mq_close (names[i]);
     mq_unlink (names[i]);
   }
}
