/********************************************
 *---- Assignment 3 -------
 * Name  : Arvind S Krishna
 * Id    : 90086152
 * Email : krishnaa@uci.edu
********************************************/

/*--- System Header files  and other include files---*/
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <pthread.h>
#include <string.h>
#include <strings.h>
#include <semaphore.h>

#define MODE 0666

int num_msg, msg_size, pairs;

/* Names of the put and take semaphores */
char ** names_put;
char ** names_take;

/*-- Function Declarations --*/
void* produce (void*);
void* consume (void*);
void print_help (void);
void unlink_semaphores (void);

/* Structure for shared memory */
typedef struct {
   char *buffer; /* Static memory pre allocated to 1000 bytes */

   sem_t *put_permits, *take_permits; 
   /* Lock that would permit only one producer or one consumer to access the buffer*/

} shared_memory_struct;

shared_memory_struct *shared_memory;

int main( int argc, char* argv[])
{
  pthread_t *threads[];
  pthread_attr_t *attr[];
  
  int counter, error, id, temp;
  int *queue_no;
  char s[1], *prefix_mem = "/my_put", *prefix_sem = "/my_take";


  /*-- Command Line option where the user enters "a.out --help" --*/
  if ( argc == 2 && strcmp (argv[1], "--help") == 0)
  {
    print_help ();
    exit(0);
  }

  /*---  Validate the input ---*/
  if (argc != 4)
  {
    printf ("\n Incorrect Argument Exception: Incorrect Number of arguments given in the command line\n");
    print_help ();
    exit(-1);
  }
  
  /*---The number of threads and wait_period is got from the command line */
  pairs = atoi (argv[1]);
  num_msg= atoi (argv[2]);
  msg_size = atoi (argv[3]);
  
  /*--- Check if they are positive integers!! ---*/
  if (pairs <= 0 || msg_size <= 0 || num_msg <= 0)
  {
     printf ("IllegalArgumentException: Negative Numbers passed as argument");
     print_help ();
     exit(-1);
  }

   queue_no    = (int*) malloc(pairs* sizeof(int));
   names_take   = (char**) malloc(pairs * sizeof(char*));
   names_put   = (char**) malloc(pairs * sizeof(char*));
   
   /* --- Create the Shared Memory ---*/
   shared_memory = (shared_memory_struct *) malloc (pairs * sizeof(shared_memory_struct));

   for (counter=0; counter < pairs; counter++)
   {
      names_take [counter] = (char*) malloc(sizeof(prefix_mem));
      names_put [counter] = (char*) malloc(sizeof(prefix_sem));

      strcpy (names_take[counter], prefix_mem);
      strcpy (names_put[counter], prefix_sem);
 
      /* --- Assign the shared memory buffer length to that of the lenght ---*/
      shared_memory[counter].buffer = (char *)malloc(msg_size * sizeof(char));

      temp = counter;
      do{
         s [0] = temp % 10 + '0';
         strcat (names_put[counter], s);
         strcat (names_take[counter], s);
         temp = temp/10;
      }while ( temp > 0);

      /* ---- Open the named semaphore -----*/
      shared_memory[counter].put_permits  = sem_open (names_put[counter], O_CREAT, MODE, 1);
      shared_memory[counter].take_permits = sem_open (names_take[counter], O_CREAT, MODE, 0);
   }

   for(counter =0; counter < 2*pairs; counter++)
   {
     
     threads[counter]   = (pthread_t*)malloc(sizeof(pthread_t));
     attr[counter]  = (pthread_attr_t*)malloc (sizeof(pthread_attr_t));

     pthread_attr_init (attr[counter]); /* initialize attr with default attributes */
     pthread_attr_setscope (attr[counter], PTHREAD_SCOPE_SYSTEM); 

     if ((counter/pairs) == 0) 
     {
       queue_no[counter] = counter;
       error = pthread_create (threads[counter], attr[counter], 
	 	               produce, (void*) &queue_no[counter]);
     }
     else
       error = pthread_create (threads[counter], attr[counter], 
	 	               consume, (void*) &queue_no[counter%pairs]);
       
     if (error != 0)
     {
       printf ("Error:Thread creation has failed %d: %d \n", error,counter);
       switch (error)
       {
          case ENOMEM: printf ("Not Enough memory in the System\n");
                       break;
          case EINVAL: printf ("value specified in the attr is invalid\n");
                       break;
          case EPERM:  printf ("The caller does not have the enough permission to create threads \n");
                       break;
	  case EAGAIN: printf ("Failed due to EAGAIN!!\n");
	               break;

	  default:     printf ("The reason why it failed is unknown\n");
       }
       exit(-1);
     }
     
   }

   /* --- Join the threads ---- */
   for(counter =0; counter < 2*pairs; counter++)
     pthread_join (*threads[counter], NULL); 

   /* ----Unlink the semaphores ---*/
   unlink_semaphores ();
    free ((void*) queue_no); 
    for (counter =0; counter < pairs; counter++)
    {
      free ((void*) names_take[counter]);
      free ((void*) names_put[counter]);
    }
    free ((void**) names_put);
    free ((void**) names_take); 
    free ((void*) shared_memory);
}

 
 void* produce (void* ptr)
 {
   int index = * (int *)ptr, i;
   char * buffer = (char*)malloc (msg_size* sizeof(char));
   
   printf ("Producer Thread %d starting \n", index);

   for (i=0; i <msg_size; i++)
    buffer[i] = 'a';

   for (i=0; i < num_msg ; i++)
   {
       /* Lock the put semaphore and if allowed write */
       sem_wait (shared_memory[index].put_permits);

       /* --- Write the data on to the shared buffer ---*/
       bcopy (buffer,shared_memory[index].buffer, msg_size);

       /* --- Awaken the Consumer ---*/
       sem_post (shared_memory[index].take_permits);
   } 
   printf ("Producer Thread %d exiting :\n", index);
   free ((void*) buffer);
 }

 void* consume (void* ptr)
 { 
   int index = * (int *)ptr, i;

   /* buffer that will be written by the producer */
   char * buffer = (char *)malloc (msg_size * sizeof(char));
     
   printf ("Consumer Thread %d starting \n", index);
   for (i=0; i < num_msg ; i++)
   {
        /* Lock the put semaphore and if allowed consume */
        sem_wait (shared_memory[index].take_permits);

        /* --- Read the data on to the private buffer from the shared buffer ---*/
         bcopy (shared_memory[index].buffer, buffer, msg_size);

         printf ("Consumer %d : 'Got message %d from Producer %d \n", index, i, index);
 
        /* --- Awaken the Producer ---*/
        sem_post (shared_memory[index].put_permits);
   }
   printf ("Consumer Thread %d exiting : \n", index);
   free ((void*) buffer);
 }

 void print_help (void)
 {
    printf ("\n\nTo run this program do the following: \n");
    printf ("---------------------------------------\n");
    printf ("1:Enter the number of pairs \n");
    printf ("2:The Number of messages that have to be sent: \n");
    printf ("3: Size of the messages that have to be sent\n");
    printf (" Entering Floating point Numbers: Integer Value will be taken in all the cases\n");
    printf (" ./a.out <pairs> <number_of_messages> <message_pairs>\n");
 }

 void unlink_semaphores (void)
 {
    int i;
    for (i =0; i < pairs; i++)
    {
       /*---- Close the semaphores --- */
       sem_close (shared_memory[i].put_permits);
       sem_close (shared_memory[i].take_permits);
       sem_unlink (names_take[i]);
       sem_unlink (names_put[i]);
    } 
 }

