/********************************************
 *---- Assignment 4 -------
 * Name  : Arvind S Krishna
 Id    : 90086152
 * Email : krishnaa@uci.edu
********************************************/

/*--- System Header files  and other include files---*/
#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>

#include "Time.h"
#include "RandomGen.h"
#define PERMS 0755


/*-- Function Declarations --*/
int file_open (char*);
int get_pid (char *);

void usr1_handler (int);
void usr2_handler (int);
void exit_program (int);
void check_if_usr1_arrived(void);


/***** External Variables ******************************/
long iterations =0;
int calibration_iterations = 5;
int wait_period = 1;
int o_pid =0, my_pid;
char *my_name;

/*
 * This global variable is incremented every time a signal 
 * is received. The type of this is sig_atomic_t mapped
 * to i_type that is mapped on to integer in most of the
 * platforms.
 */
volatile sig_atomic_t has_sig_usr1_arrived = 0;

/*******************************************************/

int main( int argc, char* argv[])
{
   int pid, fd, other_pid, counter;
   long temp_result = 0L;
   char *spid, *name1 = "sig1.pid", *name2 = "sig2.pid";
   double random;

   /* Dry run the loop for 1 sec */
   calibration_loop (wait_period);

   printf ("Computing the number of iterations to finish 1 sec: \n");

   for (counter =1; counter <= calibration_iterations; counter++)
      temp_result += calibration_loop (wait_period);

   /* --- Set the Avg number of iterations required to completed one second --*/
   iterations = temp_result / calibration_iterations;
   printf ("Loop Iterations to complete 1 sec = %d\n\n", iterations);

   /* 
    * Try and opent the first file sig1.pid if possible else
    * opent sig2.pid, if cannot open both exit.
    */ 
   if (file_open(name1) == 0)
   {
      other_pid = 2;
      my_name = name1;
   }
   else
   {
      if (file_open(name2) ==0)
      {
         other_pid =1;
         my_name = name2;
      }
      else
      {
         printf ("Error could not open either of the files!! \n");
         exit (-1);
       }
   }

   /* Set up the signal hndling stuff */
   if (signal(SIGUSR1, usr1_handler) == SIG_ERR)
   {
      perror ("signal:SIGUSR1");
      exit (1);
   }

   if (signal(SIGINT, exit_program) == SIG_ERR)
   {
       perror ("signal:SIGINT");
       exit (1);
   }

   if (signal(SIGUSR2, usr2_handler) == SIG_ERR)
   {
       perror ("Signal:SIGUSR2");
       exit (1);
   }

   /* Get the other guy's process id */
   if (other_pid == 1)
    o_pid = get_pid (name1);
   else
    o_pid = get_pid (name2);

   /* Seed the random number generator */
   random_init();

   printf ("************************************************\n");
   printf ("My    process Id = %d\n", my_pid);
   printf ("Other process Id = %d\n", o_pid);
   printf ("************************************************\n");

   /* ----- MAIN RUN LOOP -----*/
   while (1)
   {
       check_if_usr1_arrived();
       delay_loop (iterations);    

       random = get_random();
       if (random <= (double)1 && random >= (double)0)
       {
          printf ("Sending SIGUSR1 to %d\n", o_pid);
          fflush (stdout);
          kill (o_pid, SIGUSR1);
       }

       check_if_usr1_arrived();
       sleep (1);

       random = get_random();
       if (random <= (double)1 && random >= (double)0)
       {
          printf ("Sending SIGUSR1 to %d\n", o_pid);
          fflush (stdout);
          kill (o_pid, SIGUSR1);
       }
   }

}

/*
 * Locking for the file is done using the fcntl command in this function.
 * the call to lock the file is fcnlt (fd, lock, &flock).this function
 * tries to open the file speciifed file. If present it prints the pid
 * of the process in the file else and returns true else returns false. 
 */

int file_open (char * name)
{
   int lock;
   char spid[10]; 

   struct flock lck;
   
   /* set up the record locking structure, the address of which
    * is passed to the fcntl system call.
    */
   lck.l_type = F_WRLCK;   /* setting a write lock */
   lck.l_whence = 0;       /* offset l_start from beginning of file */
   lck.l_start = 0L;
   lck.l_len = 0L;         /* until the end of the file address space */
   
   int fd ; 

   /*
    * This operation is atomic if there is no file then it would
    * return an error.
    */
 
   fd = open (name, O_RDWR|O_CREAT|O_EXCL, PERMS);
   
   if (fd !=-1)
   {
      /* Try Lock the file to prevent reading and writing
       * at the same time.
       */
      lock = fcntl (fd, F_SETLK, &lck);

      /* Check if I have succeed */
      if (lock != 0)
      {
         printf ("Strange Cannot get lock on the file!\n");
         exit (0);
      }

      my_pid = getpid();
      sprintf (spid, "%d", my_pid);

      /* write the pid of this process in the file */
      write (fd, spid, strlen(spid));
      close (fd);
      fcntl (fd, F_UNLCK, &lck); 
      return 0;
    }

    return -1;
}

int get_pid (char *name)
{
  int fd = -1,pid, lock;
  struct flock lck;
  char buffer[10];

  /* Wait till the other process creates the file */ 
  while (fd == -1)
  {
     /* If I were to sleep here I could get a signal and 
      * and be interrupted hence busy-waiting is done in this part.
      */
     delay_loop(iterations);
     fd = open (name, O_RDONLY);
  }
   
   /* set up the record locking structure, the address of which
    * is passed to the fcntl system call.
    */

   lck.l_type = F_RDLCK;   /* setting a read lock */
   lck.l_whence = 0;       /* offset l_start from beginning of file */
   lck.l_start = 0L;
   lck.l_len = 0L;         /* until the end of the file address space */
      
   /* Try Lock the file */
   lock = fcntl (fd, F_SETLK, &lck);

   /* read from the file , Assuming process id cannot 
    * be more than 10 charecters long.
    */
   read (fd, buffer, 10);

   pid = atoi (buffer);

   close (fd);
   fcntl (fd, F_UNLCK, &lck); 
   return pid;
}

void usr1_handler (int param)
{
   /* --- Make the signal handler handle another occurance of the isgnal */
   signal(SIGUSR1, usr1_handler);

   /* Set that signal bit */
   /*while (has_sig_usr1_arrived);*/

   has_sig_usr1_arrived++; 
}

void usr2_handler (int param)
{
  /* Make the handler re handle the signal */
  signal (SIGUSR2, usr2_handler);
  
  /* Print that it received the signal */
  printf ("Received SIGUSR2 from %d \n", o_pid);
}


void exit_program (int param)
{
  /* print that I caught the SIGINT */
  printf ("Caught SIGINT Exiting \n");

  /* remvove its pid file from the directory */
  /*remove (my_name);*/
  unlink (my_name);

  /* Send SIGINT to the other guy and then exit */
  kill (o_pid, SIGINT);
  
  /* exit now from the program */
  exit (0);
}

void check_if_usr1_arrived()
{
   while (has_sig_usr1_arrived != 0)
   {
      /* Print that u got a signal */
      printf ("Received SIGUSR1 from %d \n", o_pid);
      
      --has_sig_usr1_arrived; 

      printf ("Sending SIGUSR2 to %d\n", o_pid);

      /* Send the signal to the other guy */
      kill (o_pid, SIGUSR2);

   }
}
      
